/*-
 * Copyright (c) 2004 - 2008 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      CTPP2Compiler.cpp
 *
 * $CTPP$
 */
#include <CTPP2Compiler.hpp>

#include <CTPP2Syntax.h>
#include <StaticData.hpp>
#include <StaticText.hpp>
#include <VMOpcodes.h>

namespace CTPP // C++ Template Engine
{

//
// Constructor
//
CTPP2Compiler::CTPP2Compiler(VMOpcodeCollector  & oIVMOpcodeCollector,
                             StaticText         & oISyscalls,
                             StaticData         & oIStaticData,
                             StaticText         & oIStaticText): oVMOpcodeCollector(oIVMOpcodeCollector),
                                                                 oSyscalls(oISyscalls),
                                                                 oStaticData(oIStaticData),
                                                                 oStaticText(oIStaticText)
{
	mSyscalls["__ctpp2_emitter"]      = oSyscalls.StoreData("__ctpp2_emitter",      15);
	mSyscalls["__ctpp2_output"]       = oSyscalls.StoreData("__ctpp2_output",       14);
	mSyscalls["__ctpp2_output_int"]   = oSyscalls.StoreData("__ctpp2_output_int",   18);
	mSyscalls["__ctpp2_output_float"] = oSyscalls.StoreData("__ctpp2_output_float", 20);
	mSyscalls["__ctpp2_output_text"]  = oSyscalls.StoreData("__ctpp2_output_text",  19);

	iZeroId = oStaticData.StoreInt(0);
	iOneId  = oStaticData.StoreInt(1);

	oVMOpcodeCollector.Insert(CreateInstruction(SYSCALL, SYSCALL_PARAMS(mSyscalls["__ctpp2_emitter"], 0), 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_SRC_STACK | ARG_DST_HR, 0, 0));
	oVMOpcodeCollector.Insert(CreateInstruction(POP, 1, 0));
}

//
// Store template source name
//
INT_32 CTPP2Compiler::StoreSourceName(CCHAR_P szName, const UINT_32 iNameLength) { return oStaticText.StoreData(szName, iNameLength); }

//
// Send variable to standard output collector
//
INT_32 CTPP2Compiler::OutputVariable(const VMDebugInfo & oDebugInfo)
{
return oVMOpcodeCollector.Insert(CreateInstruction(OUTPUT | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Send text data to standard output collector
//
INT_32 CTPP2Compiler::OutputStaticData(CCHAR_P vBuffer, const UINT_32 & iBufferLength, const VMDebugInfo & oDebugInfo)
{
	if (iBufferLength == 0) { return 0; }

	INT_32 iId = oStaticText.StoreData(vBuffer, iBufferLength);

return oVMOpcodeCollector.Insert(CreateInstruction(OUTPUT | ARG_SRC_STR, iId, oDebugInfo.GetInfo()));
}

//
// Send integer value to standard output collector
//
INT_32 CTPP2Compiler::OutputStaticData(const INT_64 & iData, const VMDebugInfo & oDebugInfo)
{
	INT_32 iId = oStaticData.StoreInt(iData);

return oVMOpcodeCollector.Insert(CreateInstruction(OUTPUT  | ARG_SRC_INT,  iId, oDebugInfo.GetInfo()));
}

//
// Send floating point value to standard output collector
//
INT_32 CTPP2Compiler::OutputStaticData(const W_FLOAT & dData, const VMDebugInfo & oDebugInfo)
{
	INT_32 iId = oStaticData.StoreFloat(dData);

return oVMOpcodeCollector.Insert(CreateInstruction(OUTPUT  | ARG_SRC_FLOAT,  iId, oDebugInfo.GetInfo()));
}

//
// Change visibility scope of CDT data (for ARRAY iterating)
//
INT_32 CTPP2Compiler::ChangeScope(CCHAR_P szScopeName, const UINT_32 & iScopeNameLength, const VMDebugInfo & oDebugInfo)
{
	UINT_64 iDebugInfo = oDebugInfo.GetInfo();
	// Remember old HR value
	// Get from DR register indirect value and store it into stack
	INT_32 iId   = oStaticText.StoreData(szScopeName, iScopeNameLength);
	INT_32 iNull = oStaticData.StoreInt(0);
//
//	PUSH     GR
//	MOVISTR  GR, HR["ScopeName"]
//	PUSH     FR
//	MOVSIZE  FR, GR.size()
//	PUSH     ER
//	MOV      ER, 0
//
//  @RET_POINT
//	CMP      ER, FR
//	JE       @END
//	MOVIREG  HR, GR[ER]
//

//
//	Loop code
//

//
//	INC      ER
//	JMP      RET_POINT
//  @END:

	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_HR                ,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_GR                ,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(MOVISTR | ARG_DST_GR | ARG_SRC_HR   ,   iId, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_FR                ,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(MOVSIZE | ARG_DST_FR | ARG_SRC_GR   ,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_ER                ,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_ER | ARG_SRC_INT  , iNull, iDebugInfo));

	UINT_32 iRetPoint = oVMOpcodeCollector.Insert(CreateInstruction(CMP     | ARG_SRC_ER | ARG_DST_FR   , iNull, iDebugInfo));

	oVMOpcodeCollector.Insert(CreateInstruction(JE                                  , (UINT_32)-1, iDebugInfo));

	oVMOpcodeCollector.Insert(CreateInstruction(MOVIREG | ARG_DST_HR | ARG_SRC_GR   , ARG_SRC_ER, iDebugInfo));

return iRetPoint;
}

//
// Reset scope to previous CDT data (for ARRAY iterating)
//
INT_32 CTPP2Compiler::ResetScope(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	UINT_64 iDebugInfo = oDebugInfo.GetInfo();
//
//	INC      ER
//	JMP      RET_POINT
//  @END:
//      POP      ER
//      POP      FR
//      POP      GR

	oVMOpcodeCollector.Insert(CreateInstruction(INC     | ARG_SRC_ER                ,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(JMP                                 ,   iIP, iDebugInfo));

	UINT_32 iEndPoint =
	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_ER | ARG_SRC_STACK,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, iDebugInfo));

	VMInstruction * pInstr = oVMOpcodeCollector.GetInstruction(iIP + 1);
	if (pInstr == NULL) { throw "Ouch!"; }
	pInstr -> argument = iEndPoint;

	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_FR | ARG_SRC_STACK,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, iDebugInfo));

	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_GR | ARG_SRC_STACK,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, iDebugInfo));

	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_HR | ARG_SRC_STACK,     0, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, iDebugInfo));

return iEndPoint;
}

//
// Get system call by id
//
INT_32 CTPP2Compiler::GetSyscallId(CCHAR_P szSyscallName, const UINT_32 & iSyscallNameLength)
{
	std::string sTMP(szSyscallName, iSyscallNameLength);
	std::map<std::string, UINT_32>::const_iterator itmSyscalls = mSyscalls.find(sTMP);
	if (itmSyscalls != mSyscalls.end()) { return itmSyscalls -> second; }

	UINT_32 iSyscallNum = oSyscalls.StoreData(szSyscallName, iSyscallNameLength);
	mSyscalls[sTMP] = iSyscallNum;

return iSyscallNum;
}

//
// Execute system call such as HREF_PARAM, FORM_PARAM, etc
//
INT_32 CTPP2Compiler::ExecuteSyscall(CCHAR_P szSyscallName, const UINT_32 & iSyscallNameLength, const UINT_32 & iArgNum, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(SYSCALL, SYSCALL_PARAMS(GetSyscallId(szSyscallName, iSyscallNameLength), iArgNum), oDebugInfo.GetInfo()));
}

//
// Push contextual variable into stack
//
INT_32 CTPP2Compiler::PushContextualVariable(const eCTPP2LoopContextVar & iTokenType, const VMDebugInfo & oDebugInfo)
{
	UINT_64 iDebugInfo = oDebugInfo.GetInfo();

	switch (iTokenType)
	{
//
//  N + 0:  MOV   AR, Zero
//  N + 1:  CMP   ER, AR
//  N + 2:  JODD  @N1
//  N + 3:  PUSH  Zero
//  N + 4:  JMP   @END
// @N1
//  N + 5:  PUSH  One
// @END
//  N + 6:  something...
//
		case TMPL_LOOP_FIRST:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, iDebugInfo));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JE                       , iIP + 4, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT       , iZeroId, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP                      , iIP + 5, iDebugInfo));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId,  iDebugInfo));
			}

//
//  N + 0:  MOV   AR, ER
//  N + 1:  INC   AR
//  N + 1:  CMP   FR, AR
//  N + 2:  JE    @N1
//  N + 3:  PUSH  Zero
//  N + 4:  JMP   @END
// @N1
//  N + 5:  PUSH  One
// @END
//  N + 6:  something...
//
		case TMPL_LOOP_LAST:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_ER, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(INC | ARG_SRC_AR             , 0, iDebugInfo));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_FR | ARG_DST_AR, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JE                       , iIP + 4, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT       , iZeroId, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP                      , iIP + 5, iDebugInfo));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId, iDebugInfo));
			}

//
//  N + 0:   MOV   AR, Zero  |
//  N + 1:   CMP   ER, AR    | 0
//  N + 2:   JE    @N0       | 1 -> 6
//  N + 5:   MOV   AR, ER    | 2
//  N + 6:   INC   AR        | 3
//  N + 7:   CMP   FR, AR    | 4
//  N + 8:   JN    @N1       | 5 -> 8
// N0:
//  N + 9:   PUSH  Zero      | 6
//  N + 10:  JMP   @END      | 7 -> 9
// @N1                       |
//  N + 5:   PUSH  One       | 8
// @END                      |
//  N + 6:  something...     | 9
//
		case TMPL_LOOP_INNER:
			{

				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, iDebugInfo));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JE                           ,  iIP + 6, iDebugInfo));

				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_ER,        0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(INC | ARG_SRC_AR             ,        0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_FR | ARG_DST_AR,        0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JN                           ,  iIP + 8, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT           ,  iZeroId, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP                          ,  iIP + 9, iDebugInfo));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT    ,   iOneId, iDebugInfo));
			}

		case TMPL_LOOP_EVEN:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, iDebugInfo));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JODD,               iIP + 4, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iZeroId, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP          ,      iIP + 5, iDebugInfo));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId, iDebugInfo));
			}

		case TMPL_LOOP_ODD:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, iDebugInfo));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JODD,               iIP + 4, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId,  iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP          ,      iIP + 5, iDebugInfo));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iZeroId, iDebugInfo));
			}

		case TMPL_LOOP_COUNTER:
			{
				// Counter
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_ER, 0, iDebugInfo));
				oVMOpcodeCollector.Insert(CreateInstruction(INC | ARG_SRC_AR             , 0, iDebugInfo));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_AR,      0, iDebugInfo));
			}

		case TMPL_LOOP_SIZE:
			{
				// Loop size
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_FR     , 0, iDebugInfo));
			}

		case TMPL_LOOP_CONTENT:
			{
				// Content of variable
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_HR     , 0, iDebugInfo));
			}
		default:
			;
	}

return 0;
}

//
// Push variable into stack
//
INT_32 CTPP2Compiler::PushVariable(CCHAR_P szVariableName, const UINT_32 & iVariableNameLength, const VMDebugInfo & oDebugInfo)
{
	UINT_64 iDebugInfo = oDebugInfo.GetInfo();

	INT_32 iId = oStaticText.StoreData(szVariableName, iVariableNameLength);

//
//  N + 0:  PUSH HR["variable"]
//  N + 1:  JE N + 3
//  N + 2:  POP 1
//  N + 3:  PUSH DR["variable"]
//  N + 4:  something...
//

	// Check from HR register indirect value
	INT_32 iPos = oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_IND_STR | ARG_DST_HR, iId, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(JE, iPos + 4, iDebugInfo));
	oVMOpcodeCollector.Insert(CreateInstruction(POP, 1, iDebugInfo));
	iPos = oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_IND_STR | ARG_DST_DR, iId, iDebugInfo));

return iPos;
}

//
// Push integer value into stack
//
INT_32 CTPP2Compiler::PushInt(const INT_64 & iVariable, const VMDebugInfo & oDebugInfo)
{
	INT_32 iId = oStaticData.StoreInt(iVariable);

return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iId, oDebugInfo.GetInfo()));
}

//
// Push Float value into stack
//
INT_32 CTPP2Compiler::PushFloat(const W_FLOAT & dVariable, const VMDebugInfo & oDebugInfo)
{
	INT_32 iId = oStaticData.StoreFloat(dVariable);

return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_FLOAT, iId, oDebugInfo.GetInfo()));
}

//
// Push string value into stack
//
INT_32 CTPP2Compiler::PushString(CCHAR_P szData, const UINT_32 & iDataLength, const VMDebugInfo & oDebugInfo)
{
	INT_32 iId = oStaticText.StoreData(szData, iDataLength);

return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_STR, iId, oDebugInfo.GetInfo()));
}

//
// Check existence of stack variable
//
INT_32 CTPP2Compiler::ExistStackVariable(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(EXIST | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Check existence of AR register variable
//
INT_32 CTPP2Compiler::ExistARReg(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(EXIST | ARG_SRC_AR, 0, oDebugInfo.GetInfo()));
}

//
// Pop variable from stack
//
INT_32 CTPP2Compiler::PopVariable(const INT_32 & iVars, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(POP, iVars, oDebugInfo.GetInfo()));
}

//
// Remove last instruction from code segment
//
INT_32 CTPP2Compiler::RemoveInstr() { return oVMOpcodeCollector.Remove(); }

//
// Unconditional jump
//
INT_32 CTPP2Compiler::UncondJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JMP, iIP, oDebugInfo.GetInfo()));
}

//
// Jump if NOT Equal
//
INT_32 CTPP2Compiler::NEJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_NE, iIP, oDebugInfo.GetInfo()));
}

//
// Jump if Equal
//
INT_32 CTPP2Compiler::EQJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_EQ, iIP, oDebugInfo.GetInfo()));
}

//
// Jump if Greater
//
INT_32 CTPP2Compiler::GTJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_GT, iIP, oDebugInfo.GetInfo()));
}

//
// Jump if Greater Or Equal
//
INT_32 CTPP2Compiler::GEJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_EQ | FL_GT, iIP, oDebugInfo.GetInfo()));
}

//
// Jump if Less
//
INT_32 CTPP2Compiler::LTJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_LT, iIP, oDebugInfo.GetInfo()));
}

//
// Jump if Less Or Equal
//
INT_32 CTPP2Compiler::LEJump(const UINT_32 & iIP, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_EQ | FL_LT, iIP, oDebugInfo.GetInfo()));
}

//
// Clear register
//
INT_32 CTPP2Compiler::ClearStack(const UINT_32 & iStackPointer, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(CLEAR | ARG_DST_STACK, iStackPointer, oDebugInfo.GetInfo()));
}

//
//  Stop execution
//
INT_32 CTPP2Compiler::Halt(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(HLT, 0, oDebugInfo.GetInfo()));
}

//
// Addition
//
INT_32 CTPP2Compiler::OpAdd(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(ADD | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Substraction
//
INT_32 CTPP2Compiler::OpSub(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(SUB | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Multiplication
//
INT_32 CTPP2Compiler::OpMul(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(MUL | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Division
//
INT_32 CTPP2Compiler::OpDiv(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(DIV | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Integer division
//
INT_32 CTPP2Compiler::OpIDiv(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(IDIV | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Mod
//
INT_32 CTPP2Compiler::OpMod(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(MOD | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Negation
//
INT_32 CTPP2Compiler::OpNeg(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(NEG | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Logical negation
//
INT_32 CTPP2Compiler::OpNot(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(NOT | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Comparison
//
INT_32 CTPP2Compiler::OpCmp(const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_DST_STACK | ARG_SRC_STACK, 0, oDebugInfo.GetInfo()));
}

//
// Get last instruction number
//
UINT_32 CTPP2Compiler::GetCodeSize() const { return oVMOpcodeCollector.GetCodeSize(); }

//
// Push variable on stack into stack
//
INT_32 CTPP2Compiler::PushStackVariable(const INT_32 & iStackPos, const VMDebugInfo & oDebugInfo)
{
	return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_STACK, iStackPos, oDebugInfo.GetInfo()));
}

//
// Get instrcution by instruction number
//
VMInstruction * CTPP2Compiler::GetInstruction(const UINT_32 & iIP) { return oVMOpcodeCollector.GetInstruction(iIP); }

//
// A destructor
//
CTPP2Compiler::~CTPP2Compiler() throw()
{
	;;
}

} // namespace CTPP
// End.
